use std::process;

#[derive(Default)]
struct CPUFlags {
    overflow: bool,
    sub: bool,
}

pub struct CPU {
    prgmem: [u8; 0xFFFF + 1],
    memory: [u8; 0xFFFF + 1],
    prc: u16,
    acc: u8,
    irx: u8,
    iry: u8,
    pub is_on: bool,
    flags: CPUFlags,
}

impl CPU {
    //FIXME: pass in debug info here
    pub fn from_prg_vec(program: Vec<u8>) -> Self {
        let mut prgmem = [0; 0xFFFF + 1];
        for (idx, ele) in program.iter().enumerate() {
            prgmem[idx] = *ele;
        }
        CPU {
            prgmem: prgmem,
            memory: [0; 0xFFFF + 1],
            prc: 0,
            acc: 0,
            irx: 0,
            iry: 0,
            is_on: true,
            flags: CPUFlags::default(),
        }
    }

    fn get_next_8bit_value(&mut self) -> u8 {
        self.prgmem[{
            self.prc += 1;
            self.prc
        } as usize]
    }

    fn get_next_16bit_value(&mut self) -> u16 {
        256 * self.prgmem[{
            self.prc += 1;
            self.prc
        } as usize] as u16
            + self.prgmem[{
                self.prc += 1;
                self.prc
            } as usize] as u16
    }

    fn write_to_serial(&mut self) {
        // FIXME: make serial offset configurable
        if self.memory[0x010F + 8] > 0 {
            let mut charcode = 0;
            // FIXME: make this less jank. bit manip maybe?
            if self.memory[0x010F] > 0 {
                charcode += 64;
            }
            if self.memory[0x0110] > 0 {
                charcode += 32;
            }
            if self.memory[0x0111] > 0 {
                charcode += 16;
            }
            if self.memory[0x0112] > 0 {
                charcode += 8;
            }
            if self.memory[0x0113] > 0 {
                charcode += 4;
            }
            if self.memory[0x0114] > 0 {
                charcode += 2;
            }
            if self.memory[0x0115] > 0 {
                charcode += 1;
            }
            if self.memory[0x0118] > 0 {
                charcode = self.memory[0x0118];
            }
            if charcode == 0 && self.memory[0x0116] == 0 {
                return;
            }

            if self.memory[0x0119] == 0 {
                print!("{}", char::from_u32(charcode.into()).unwrap());
            } else {
                print!("{}", charcode)
            }
            if self.memory[0x0116] > 0 {
                println!("")
            }
        }
    }

    // NOTE: most opcode implementations are translated from C, in all their janky glory
    pub fn step(&mut self) {
        let inst = self.prgmem[self.prc as usize];
        match inst {
            0x00 => {}
            0x01 => self.memory[self.get_next_16bit_value() as usize] += 1,
            0x02 => self.acc += 1,
            0x03 => self.irx += 1,
            0x04 => self.iry += 1,
            0x05 => self.memory[self.get_next_16bit_value() as usize] -= 1,
            0x06 => self.acc -= 1,
            0x07 => self.irx -= 1,
            0x08 => self.iry -= 1,
            0x10 => {
                let new_prc = self.get_next_16bit_value();
                let addr1 = self.get_next_16bit_value() as usize;
                let addr2 = self.get_next_16bit_value() as usize;
                if self.memory[addr1] == self.memory[addr2] {
                    self.prc = new_prc - 1
                }
            }
            0x12 => {
                let new_prc = self.get_next_16bit_value();
                let addr1 = self.get_next_16bit_value() as usize;
                let addr2 = self.get_next_16bit_value() as usize;
                if self.memory[addr1] != self.memory[addr2] {
                    self.prc = new_prc - 1
                }
            }
            0x14 => {
                // JIS; Jump If SUB (😳)
                let new_prc = self.get_next_16bit_value();
                if self.flags.sub {
                    self.flags.sub = false;
                    self.prc = new_prc - 1;
                }
            }
            0x15 => {
                //JGT; Jump If Greater Than
                let new_prc = self.get_next_16bit_value();
                let addr1 = self.get_next_16bit_value() as usize;
                let addr2 = self.get_next_16bit_value() as usize;
                if self.memory[addr1] > self.memory[addr2] {
                    self.prc = new_prc - 1
                }
            }
            0x16 => {
                //JGT; Jump If Less Than
                let new_prc = self.get_next_16bit_value();
                let addr1 = self.get_next_16bit_value() as usize;
                let addr2 = self.get_next_16bit_value() as usize;
                if self.memory[addr1] < self.memory[addr2] {
                    self.prc = new_prc - 1
                }
            }
            0x1A => {
                let new_prc = self.get_next_16bit_value();
                self.prc = new_prc - 1
            }
            0x20 => self.irx = self.memory[self.get_next_16bit_value() as usize],
            0x23 => self.iry = self.memory[self.get_next_16bit_value() as usize],
            0x26 => {
                let addr1 = self.get_next_16bit_value() as usize;
                let addr2 = self.get_next_16bit_value() as usize;
                self.memory[addr2] = self.memory[addr1]
            }
            0x27 => self.memory[self.get_next_16bit_value() as usize] = self.get_next_8bit_value(),
            0x2B => self.memory[self.get_next_16bit_value() as usize] = self.iry,
            0x2C => self.memory[self.get_next_16bit_value() as usize] = self.acc,
            0x42 => {
                // ADD absolute
                let addr1 = self.get_next_16bit_value() as usize;
                let addr2 = self.get_next_16bit_value() as usize;
                self.acc = self.memory[addr2] + self.memory[addr1]
            }
            0x47 => {
                // SUB absolute
                let addr1 = self.get_next_16bit_value() as usize;
                let addr2 = self.get_next_16bit_value() as usize;
                let val1: i16 = self.memory[addr1].into();
                let val2: i16 = self.memory[addr2].into();
                let tmp_val = val1 - val2;
                if tmp_val < 0 {
                    self.flags.sub = true
                }
                if tmp_val < -255 || tmp_val > 255 {
                    self.flags.overflow = true
                }
                self.acc = tmp_val as u8;
            }
            0x50 => todo!(), // FIXME: wtf does ~(irx & iry) mean
            0xFC => self.is_on = false,
            0xFF => {}
            _ => {
                eprintln!("Opcode 0x{:x} not implemented!", inst);
                process::exit(1);
            }
        }
        self.prc += 1;

        // FIXME: serial io peripheral should be optional
        self.write_to_serial();
    }
}
